import pickle
import numpy as np
import matplotlib.pyplot as plt
from cannon.dataset import Dataset
from cannon.helpers import Table
from matplotlib import rc
import matplotlib.gridspec as gridspec
import pickle

plt.rc('text', usetex=True)
plt.rc('font', family='serif')

test_SNR = pickle.load(open('test_SNR.p', 'r'))
test_labels = pickle.load(open('test_labels.p', 'r'))
label_file = "example_DR12/apogee_test_labels.csv"
data = Table(label_file)
data.sort('id')
label_names = data.keys()
nlabels = len(label_names)
apogee_label_vals = np.array([data[k] for k in label_names]).T

names = ['T_{eff}', '\log g', '[Fe/H]', '[\\alpha/Fe]'] 

for i in range(0, len(names)):
    name = names[i]
    cannon = np.array(test_labels[:,i])
    orig = np.array(apogee_label_vals[:,i+1], dtype=float)
    snr = test_SNR
    name = names[i]
    bad = orig < -8000
    good = snr > 50
    orig = np.ma.array(orig, mask=bad)
    cannon = np.ma.array(cannon, mask=bad)
    snr = np.ma.array(snr, mask=bad)
    orig = orig[good]
    cannon = cannon[good]
    snr = snr[good]

    scatter = np.round(np.std(orig-cannon),3)
    bias  = np.round(np.mean(orig-cannon),4)
    low = np.minimum(min(orig), min(cannon))
    high = np.maximum(max(orig), max(cannon))

    fig = plt.figure(figsize=(10,6))
    gs = gridspec.GridSpec(1,2,width_ratios=[2,1], wspace=0.3)
    ax1 = plt.subplot(gs[0])
    ax2 = plt.subplot(gs[1])
    ax1.plot([low, high], [low, high], 'k-', linewidth=2.0, label="x=y")
    ax1.set_xlim(low, high)
    ax1.set_ylim(low, high)
    ax1.legend(fontsize=14)
    pl = ax1.scatter(orig, cannon, marker='x', c=snr, 
            vmin=50, vmax=200, alpha=0.7)
    cb = plt.colorbar(pl, ax=ax1, orientation='horizontal')
    cb.set_label('SNR from LAMOST Spectra', fontsize=12)
    cb.ax.tick_params(labelsize=14)
    textstr = 'Scatter: %s \nBias: %s' %(scatter, bias)
    ax1.text(0.05, 0.95, textstr, transform=ax1.transAxes, 
            fontsize=14, verticalalignment='top')
    ax1.tick_params(axis='x', labelsize=14)
    ax1.tick_params(axis='y', labelsize=14)
    ax1.set_xlabel("APOGEE Label", fontsize=14)
    ax1.set_ylabel("Cannon-LAMOST Label", fontsize=14)
    ax1.set_title("Cannon-LAMOST Output vs. APOGEE Values \n Label $%s$ for SNR $\geq$ 50" % name, fontsize=14)
    diff = cannon - orig
    npoints = len(diff)
    mu = mean(diff)
    sig = std(diff)
    ax2.hist(diff, range=[-3*sig,3*sig], color='k', bins=np.sqrt(npoints), 
            orientation='horizontal', alpha=0.3, histtype='stepfilled')
    ax2.tick_params(axis='x', labelsize=14)
    ax2.tick_params(axis='y', labelsize=14)
    ax2.set_xlabel("Count", fontsize=14)
    ax2.set_ylabel("Difference", fontsize=14)
    ax2.axhline(y=0, c='k', lw=3, label='Difference=0')
    ax2.set_title("Cannon-LAMOST Output Minus \n APOGEE Labels for $%s$" %name, 
            fontsize=14)
    ax2.legend(fontsize=14)
    plt.show()
    plt.savefig('1to1_%s.png'%i)

# first-order coeffs plot
lams = pickle.load(open("wl.p", "r"))
coeffs_all = pickle.load(open("coeffs_all.p", "r"))
scatters = pickle.load(open("scatters.p", "r"))
bad = scatters < 0.0002
scatters = np.ma.array(scatters, mask=bad)
lams = np.ma.array(lams, mask=bad)
cut = 40

lams = lams[cut:]
scatters = scatters[cut:]

nlabels = 4
fig, axarr = plt.subplots(5, figsize=(8,8), sharex=True)
#ax = plt.gca()
#ax.xaxis.grid(True)
plt.subplots_adjust(hspace=0.001)
nbins = len(ax1.get_xticklabels())
for i in range(1,5):
    axarr[i].yaxis.set_major_locator(MaxNLocator(nbins=nbins, prune='upper'))
plt.xlabel(r"Wavelength $\lambda (\AA)$", fontsize=14)
plt.xlim(np.ma.min(lams), np.ma.max(lams))
plt.tick_params(axis='x', labelsize=14)
axarr[0].set_title(
    "First-Order Fit Coefficients and Scatter from the Spectral Model",
    fontsize=14)
ax.locator_params(axis='x', nbins=10)

for i in range(0,4):
    ax = axarr[i]
    lbl = r'$%s$'%names[i]
    ax.set_ylabel(lbl, fontsize=14)
    ax.tick_params(axis='y', labelsize=14)
    ax.xaxis.grid(True)
    y = np.ma.array(coeffs_all[:,i+1], mask=bad)
    y = y[cut:]
    ax.step(lams, y, where='mid', linewidth=0.5, c='k')
    ax.locator_params(axis='y', nbins=4)

ax = axarr[4] 
ax.tick_params(axis='y', labelsize=14)
ax.set_ylabel("scatter", fontsize=14)
top = np.max(scatters[scatters < 0.8])
stretch = np.std(scatters[scatters < 0.8])
ax.set_ylim(0, top + stretch)
ax.step(lams, scatters, where='mid', c='k', linewidth=0.7)
ax.xaxis.grid(True)
ax.locator_params(axis='y', nbins=4)
fig.savefig('leading_coeffs.png')



from __future__ import (absolute_import, division, print_function)
from lamost import load_spectra, load_labels
from cannon.dataset import Dataset
from cannon.model import CannonModel
from cannon.spectral_model import draw_spectra, diagnostics, triangle_pixels, overlay_spectra, residuals
import numpy as np
import pickle
import random
import csv
from matplotlib import rc

plt.rc('text', usetex=True)
plt.rc('font', family='serif')

# STEP 0: CHOOSE TRAINING SET

allfiles = np.loadtxt("example_LAMOST/lamost_sorted_by_ra.txt", dtype=str)
dir_dat = "example_LAMOST/Data_All"
tr_files = allfiles
wl, tr_flux, tr_ivar = load_spectra(dir_dat, tr_files)
SNR_raw = tr_flux * tr_ivar**0.5
bad = SNR_raw == 0
SNR_raw = np.ma.array(SNR_raw, mask=bad)
SNR = np.ma.median(SNR_raw, axis=1)
cut = 130
tr_files = allfiles[SNR>cut]
outputf = open("example_LAMOST/tr_files.txt", "w")
for tr_file in tr_files: 
    outputf.write(tr_file + '\n')
outputf.close()
test_files = set(allfiles) - set(tr_files)
test_files = np.array(test_files)
outputf = open("example_LAMOST/test_files.txt", "w")
for test_file in test_files:
    outputf.write(test_file + '\n')
outputf.close()
all_ids, all_tr_labels = load_labels(dir_lab)
tr_labels = np.zeros((len(tr_files), all_tr_labels.shape[1]))
for jj,tr_id in enumerate(tr_files):
    tr_labels[jj,:] = all_tr_labels[all_ids==tr_id,:]
pickle.dump(tr_labels, open('tr_labels.p', 'w'))

# STEP 1: PREPARE DATA 
if glob.glob('lamost_data.p'):
    wl, tr_flux, tr_ivar, tr_label, test_flux, test_ivar = pickle.load(
            open('lamost_data.p', 'r'))

else:
    #tr_files = np.genfromtxt("example_LAMOST/Training_Data.txt", dtype=str)
    #test_files = np.loadtxt("example_LAMOST/Test_Data.txt", dtype=str)
    dir_lab = "example_DR12/reference_labels.csv"
    dir_dat = "example_LAMOST/Data_All"

    wl, tr_flux, tr_ivar = load_spectra(dir_dat, tr_files)
    test_flux = tr_flux
    test_ivar = tr_ivar
    #wl, test_flux, test_ivar = load_spectra(dir_dat, test_files)
    tr_label = pickle.load(open('tr_labels.p', 'r'))

    pickle.dump((wl, tr_flux, tr_ivar, tr_label, test_flux, test_ivar), 
            open('lamost_data_1to1.p', 'w'))

dataset = Dataset(wl, tr_flux, tr_ivar, tr_label, test_flux, test_ivar)

# set the headers for plotting
dataset.set_label_names(['T_{eff}', '\log g', '[M/H]', '[\\alpha/Fe]'])

# Plot SNR distributions and triangle plot of reference labels
dataset.diagnostics_SNR()
dataset.diagnostics_ref_labels()


# STEP 2: CONTINUUM IDENTIFICATION

# Pseudo-continuum normalization for the training spectra
if glob.glob('pseudo_normed_spec.p', 'r'):
    norm_tr_fluxes, norm_tr_ivars = pickle.load(open("pseudo_normed_spec.p", "r"))

else:
    norm_tr_fluxes, norm_tr_ivars = dataset.continuum_normalize_q(
            dataset.tr_flux, dataset.tr_ivar, q=0.90, delta_lambda=400)
    pickle.dump((norm_tr_fluxes, norm_tr_ivars), 
            open("pseudo_normed_spec.p", "w"))

# From the cont norm training spectra, identify continuum pixels
if glob.glob('contmask.p', 'r'):
    contmask = pickle.load(open("contmask.p", "r"))
else:
    # Identify the best 5% of continuum pixels
    # contmask = dataset.make_contmask(norm_tr_fluxes, norm_tr_ivars, frac=0.05)

    # Identify the best 5% of continuum pixels in each of the following
    # pixel regions 
    dataset.ranges = [[0,50], [50,100], [100,400], [400,600], [600,1722], [1863, 1950], [1950, 2500], [2500,3000], [3000, len(dataset.wl)]]
    contmask = dataset.make_contmask(norm_tr_fluxes, norm_tr_ivars, frac=0.05)

dataset.set_continuum(contmask)


# RUN CONTINUUM NORMALIZATION CODE

if glob.glob('cont.p', 'r'):
    tr_cont, test_cont = pickle.load(open("cont.p", "r"))
else:
    dataset.ranges = [[0,1723], [1863,len(dataset.wl)]] # split into two wings
    tr_cont, test_cont = dataset.fit_continuum(deg=3, ffunc="sinusoid")
    pickle.dump((tr_cont, test_cont), open("cont.p", "w"))


# Check out the median flux overlaid with cont pix 
f_bar = np.zeros(len(dataset.wl))
sigma_f = np.zeros(len(dataset.wl))
for wl in range(0,len(dataset.wl)):
    flux = norm_tr_fluxes[:,wl]
    ivar = norm_tr_ivars[:,wl]
    f_bar[wl] = np.median(flux[ivar>0])
    sigma_f[wl] = np.sqrt(np.var(flux[ivar>0]))
bad = np.var(norm_tr_ivars, axis=0) == 0
f_bar = np.ma.array(f_bar, mask=bad)
sigma_f = np.ma.array(sigma_f, mask=bad)
plot(dataset.wl, f_bar, alpha=0.7)
fill_between(dataset.wl, (f_bar+sigma_f), (f_bar-sigma_f), alpha=0.2)
scatter(dataset.wl[contmask], f_bar[contmask], c='r')
xlim(3800,9100)
ylim(0,1.1)
xlabel("Wavelength (A)")
ylabel("Median Flux")
title("Median Flux Across Training Spectra Overlaid with Cont Pix")
savefig("medflux_contpix.png")

# Residuals between raw flux and fitted continuum
res = (dataset.tr_flux-tr_cont)*np.sqrt(dataset.tr_ivar)
# sort by temperature...
sorted_res = res[np.argsort(dataset.tr_label[:,2])]
im = plt.imshow(res, cmap=plt.cm.bwr_r, interpolation='nearest', 
        vmin = -80, vmax=10, aspect = 'auto', origin = 'lower',
        extent=[0, len(dataset.wl), 0, nstars])
xlabel("Pixel")
ylabel("Training Object")
title("Residuals in Continuum Fit to Raw Spectra")
colorbar()


# Plot the cont fits for 100 random training stars
pickstars = np.zeros(100)
nstars = dataset.tr_flux.shape[0]
for jj in range(100):
    pickstars[jj] = random.randrange(0, nstars-1)

for jj in pickstars:
    #bad = np.var(norm_tr_ivars, axis=0) == 0
    bad = norm_tr_ivars[jj,:] == 0 
    flux = np.ma.array(dataset.tr_flux[jj,:], mask=bad)
    plot(dataset.wl, flux, alpha=0.7)
    scatter(dataset.wl[contmask], flux[contmask], c='r')
    cont = np.ma.array(tr_cont[jj,:], mask=bad)
    plot(dataset.wl, cont)
    xlim(3800, 9100)
    title("Sample Continuum Fit")
    xlabel("Wavelength (A)")
    ylabel("Raw Flux")
    savefig('contfit_%s.png' %jj)
    plt.close()

f,ax = plt.subplots(2, sharex=True)
meanres = np.mean(res, axis=0)
mu = mean(meanres)
sig = np.sqrt(var(meanres))
c= ax[1].imshow(sorted_res, cmap = plt.cm.bwr_r, interpolation='nearest',
     vmin = mu-3*sig, vmax= mu+sig, aspect='auto')
colorbar(c, orientation='horizontal')
ax[0].plot(flux, alpha=0.7)
ax[0].plot(cont)
xlim(0,3626)

ax[0].hist(meanres, bins = 60, range=(mu-3*sig, mu+sig))
ax[0].set_ylabel("Raw Flux")
title("Sample Raw Spec with All Continuum Fit Residuals")
c = ax[1].imshow(res, cmap = plt.cm.bwr_r, interpolation='nearest',
        vmin = -80, vmax=10, aspect='auto')
colorbar(c, orientation='horizontal')
ax[1].set_ylabel("Training Objects")


norm_tr_fluxes, norm_tr_ivars, norm_test_fluxes, norm_test_ivars = \
        dataset.continuum_normalize_f(cont=(tr_cont, test_cont))


# If you approve...

dataset.tr_flux = norm_tr_fluxes
dataset.tr_ivar = norm_tr_ivars
dataset.test_flux = norm_test_fluxes
dataset.test_ivar = norm_test_ivars

# learn the model from the reference_set
model = CannonModel(dataset, 3) # 2 = quadratic model
model.fit() # model.train would work equivalently.

# or...
coeffs_all = pickle.load(open("coeffs_all.p", "r"))

# check the model
model.diagnostics()

# infer labels with the new model for the test_set
if glob.glob('test_labels.p'):
    test_label = pickle.load(open('test_labels.p', 'r'))
    dataset.test_label = test_label
else:
    dataset, label_errs = model.infer_labels(dataset)

# Make plots
dataset.dataset_postdiagnostics(dataset)

cannon_set = draw_spectra(model.model, dataset)
diagnostics(cannon_set, dataset, model.model)
